c  Small working utility: for a given equatorial coordinates
c  al,de (or galactic l,b), radial velocity RV and 3D box size SX,SX,SV
c  (pixels, or degrees and km/s)
c  it finds the best CGPS mosaic field (if any), and generates
c  a shell script file for the subcube extraction.

c  Requires: file "fields" where each line (starting from the second one)
c  contains the following information about one mosaic field:

c  field file  NX    NY    NV   CRVAL1 CRPIX1 CDELT1  CRVAL2 CRPIX2 CDELT2   CRVAL3 CRPIX3     CDELT3

c  A comment: file name should contain a full path from the CDROM mounting
c  point. If the name contains "/" or any other special symbols, it should be
c  enclosed in quotes.

c  Current limitations:
c  - Only square boxes (SX=SY);
c  - Only |CDELT1| = |CDELT2| fields (the rest are ignored);
c  - Mosaics should be in the galactic frame of reference;
c  - Only cases when the box is completely within the field boundaries.
c    (Which also means that if the spatial box size is > 224 pix., then
c     there's a chance that a good field won't be found - even if the area
c     is fully covered by the CGPS. Future versions will try to do the
c     job by cutting subcubes from ajacent fields (up to 4), and then
c     gluing them together.)


c  1.01 - More friendly user interface.
c  1.02 - Script shell generation
c  1.03 - Minor bugs corrected. Added B1950 equatorial equinox.
c  1.04 - A bug with subfits arguments was corrected. Simpler interface.
c  1.05 - Fixed bug for dV=1.

c  Author: Sergey Mashchenko (2000,2001)

c  Any comments - to syam@astro.umontreal.ca, or Mashchenko@excite.com

      program find_field
      implicit NONE
      integer NF_max, I,J,strlen_trail
c  Maximum number of fields:
      parameter (NF_max=1000)
      real*8  Rval(3,NF_max),Rpix(3,NF_max),Delt(3,NF_max),al,de,RV,
     #        Pi,l,b,X(3),Xa,Xb,Numbers(3),Alpha,Delta,dX,dV,
     #        SXt,SVt,Equinox
      integer N(3,NF_max), NF,SX1,SX2,SV1,SV2,S1(3,NF_max),
     #        S2(3,NF_max),Ibest,SX,SV,fl,
     #        Xa_best(3),Xb_best(3),nn,strlen_lead
      character*3 Field(NF_max), Coord*1, Units*1, String*256,Coords*256
      character*256 File(NF_max)
      logical ok(NF_max), Gal, runit

      write (*,*) 
      write (*,*) 
     #  '   *********  Program Find_field, ver. 1.05  *********'
      write (*,*)

      Pi = dacos(-1.d0)

c                    Reading inputs:

 3    continue
      write (*,*) ' Type of coordinates (\"g\" - galactic, \"e1950\" -'
     #            //' equatorial B1950,'
      write (*,*) '   and \"e2000\" - equatorial J2000):'
      read (*,'(A)') Coords
      Coords = Coords(len(Coords)-strlen_lead(Coords)+1:len(Coords))
      Coord = Coords(1:1)
      Gal = (Coord.eq.'g' .or. Coord.eq.'G')
      if (.not.Gal .and. Coord.ne.'e'.and. Coord.ne.'E')  then
         write (*,*) ' Wrong type of coordinates!'
         write (*,*)
         go to 3
         end if
      if (Gal)  then
         write (*,*) '     Galactic coordinates:'
         write (*,*) 'l (dgr.):'
         read (*,*) l
         write (*,*) ' b (dgr.):'
         read (*,*) b
        else
         read(Coords(2:len(coords)),*)  Equinox
         write (*,'(A,F6.1,A)') '     Equatorial coordinates (',
     #                          Equinox,'):'
 5       continue
         write (*,*) 'RA:'
         write (*,*) '  (Accepted formats: \"hh mm ss.s\", \"hh mm.m\",'
     #               //' \"dgr\")'
         read (*,'(A)') String
         call Extract_numbers (String, nn,Numbers)
         if (nn .eq. 0)  go to 5
         al = Alpha (Numbers,nn)

 6       continue
         write (*,*) 'Dec:' 
         write (*,*) '  (Accepted formats: \"dd mm ss.s\", \"dd mm.m\",'
     #               //' \"dgr\")'
         read (*,'(A)') String
         call Extract_numbers (String, nn,Numbers)
         if (nn .eq. 0)  go to 6
         de = Delta (Numbers,nn)
        end if
      write (*,*) 'RV (km/s):'
      read (*,*) RV

 4    continue
      write (*,*) 'Box size dX,dV and units (\"p\" for pixels, '//
     #    '\"r\" for real - dgr. for dX, and km/s for dV; e.g. - '//
     #    '"50 20 p")'
      write (*,*) '  (dV=0 means that all the range of V will be'//
     #            ' extracted)'
      read (*,*) SXt,SVt,Units
      runit = (Units.eq.'r' .or. Units.eq.'R')
      if (runit)  then
         dX = SXt
         dV = SVt
         end if
      if (.not.runit .and. Units.ne.'p'.and. Units.ne.'P')  then
         write (*,*) ' Wrong units!'
         write (*,*)
         go to 4
         else
         SX = SXt
         SV = SVt
         end if
            
      open (1,file='fields',status='old')
      read (1,*)
      I = 0
 1    continue
      I = I + 1
      if (I .gt. NF_max)  then
         write (*,*) ' Too small NF_max parameter!'
         stop
         end if
      read (1,*,end=2) Field(I),File(I),N(1,I),N(2,I),N(3,I),(Rval(J,I),
     #    Rpix(J,I),Delt(J,I),J=1,3)
      if (dabs(Delt(1,I)) .ne. dabs(Delt(2,I)))  then
         write (*,*) 'CDELT1 should be = CDELT2; field ',
     #               Field(I),' has been skipped'
         I = I - 1
         go to 1
         end if

      if (runit)  then
         SX = dnint(dX / dabs(Delt(1,I)))
         if (SX .lt. 1)  SX=1
         SV = dnint(dV / dabs(Delt(3,I)))
         if (SV .lt. 1)  SV=1
         end if

      if (mod(SX,2).eq.0)  then
         SX1 = SX/2
         SX2 = SX1-1
        else
         SX1 = (SX-1)/2
         SX2 = SX1
        end if
      if (mod(SV,2).eq.0)  then
         SV1 = SV/2
         SV2 = SV1-1
        else
         SV1 = (SV-1)/2
         SV2 = SV1
        end if
c  Pixel coordinates of the box boundaries (0 corresponds to the box center):
      S1(1,I) = SX1
      S1(2,I) = SX1
      if (SV .ne. 0) then
         S1(3,I) = SV1
         S2(3,I) = SV2
        else
         S1(3,I) = -1
         S2(3,I) = -1
        end if
      S2(1,I) = SX2
      S2(2,I) = SX2
      go to 1

 2    continue
      NF = I - 1

c  Transforming equatorial into galactic coordinates:      
      if (.not.Gal)  then
         write (*,*) al, de
         al = al/180.d0*Pi
         de = de/180.d0*Pi
         call galactic_equatorial (.false.,Equinox,l,b,al,de)
         l = l*180.d0/Pi
         b = b*180.d0/Pi
         end if

      write (*,*) 
      write (*,'(A,2F11.4)') 'Galactic coordinates l,b (dgr.):', l, b
      write (*,*) 

      Ibest = 0
      do I=1,NF
         X(1) = (l-Rval(1,I))/Delt(1,I) + Rpix(1,I)
         X(2) = (b-Rval(2,I))/Delt(2,I) + Rpix(2,I)
         X(3) = (RV-Rval(3,I))/Delt(3,I) + Rpix(3,I)
         ok(I) = .true.
         do J=1,3
            if (J.ne.3.or.S1(J,I).ne.-1)  then
               Xa = X(J) - S1(J,I)
               Xb = X(J) + S2(J,I)
              else
               Xa = 1
               Xb = N(J,I)
              end if
            if (Xa.lt.1 .or. Xb.gt.N(J,I))  then
c  We reject the field, if at least one side of our box is outside of the field boundaries:
               ok(I) = .false.
              end if
            end do
c  Any field which completely contains the box is considered to be equally good:
         if (ok(I))  then
            do J=1,3               
               Ibest = I
               if (J.ne.3.or.S1(J,I).ne.-1)  then
                  Xa_best(J) =  X(J) - S1(J,I)
                  Xb_best(J) =  X(J) + S2(J,I)
                 else
                  Xa_best(J) =  1
                  Xb_best(J) =  N(J,I)
                 end if
               end do
            end if
         end do

      open (1,file='script.sh')
      if (Ibest .ne. 0)  then
         fl = strlen_trail(File(IBest))
         write (*,*) ' The best field: ', Field(Ibest)
         write (*,*)
         write (1,'(A)') '#!/bin/sh'
c  For different OSTYPE variable values one should specify a mounting
c   point for the cdrom (CD_DIR) and whether one should mount/unmount cdrom
c   every time (MOUNT=1); currently only linux and solaris are specified. 
c   Command "eject $CD_ROM" should eject CDROM.
c   For Linux the program relies on information in /etc/fstab file.
         write (1,'(A)') 'OS_TYPE=`uname`'
         write (1,'(A)') 'if [ ${OS_TYPE} = \'Linux\' ]'
         write (1,'(A)') '  then   \\rm _fstab; cat /etc/fstab|grep -i'
     #                 //' iso9660>_fstab; read _A _B _C<_fstab'
         write (1,'(A)') '         \\rm _fstab'
         write (1,'(A)') '       CD_DIR=${_B}; MOUNT=1;'//
     #                   ' CD_ROM=${_B}'
         write (1,'(A)') '  elif [ ${OS_TYPE} = \'SunOS\' ]'
         write (1,'(A)') '    then CD_DIR=/cdrom/cdrom0; MOUNT=0;'//
     #                   ' CD_ROM=cdrom'
         write (1,'(A)') '  else echo Unknown OS_TYPE ${OS_TYPE} \\!;'//
     #                   ' exit'
         write (1,'(A)') '  fi'
         write (1,'(A)') '_BUSY_=0'
         write (1,'(A)') 'until'
         write (1,'(A)') '    if [ ${_BUSY_} = 0 ]; then echo '//
     #                  'Please insert CDROM with the field '//
     #                  Field(Ibest)//'\\; press \\<Enter\\> '//
     #                  'when ready; fi'
         write (1,'(A)') '    read _KKK_'
         write (1,'(A)') '    if [ ${MOUNT} = 1 ]; then mount'//
     #                   ' ${CD_DIR}; fi'
         write (1,'(A)') '    [ -f ${CD_DIR}/'//File(Ibest)(1:fl)//' ]'
         write (1,'(A)') '  do'
         write (1,'(A)') '    if [ ${MOUNT} = 0 -a ! -d ${CD_DIR} ]'
         write (1,'(A)') '      then'
         write (1,'(A)') '        echo Device is busy\\!'//
     #                   ' Press \\<Enter\\> again when it is ready.'
         write (1,'(A)') '        _BUSY_=1'
         write (1,'(A)') '      else'
         write (1,'(A)') '        echo Wrong CD\\! No such file: '//
     #                   '${CD_DIR}/'//File(Ibest)(1:fl)
         write (1,'(A)') '        eject ${CD_DIR}'
         write (1,'(A)') '        _BUSY_=0'
         write (1,'(A)') '      fi'
         write (1,'(A)') '  done'
         write (1,'(A,6I6)') '${OS_TYPE}/subfits ${CD_DIR}/'//
     #                  File(Ibest)(1:fl)//' _subcube1.fits ',
     #                  (Xa_best(J),J=1,3), (Xb_best(J),J=1,3)
         write (1,'(A)') 'eject ${CD_ROM}'
         else
         write (*,*) ' There is no good field!'
         write (*,*)
         end if
      close (1)

      stop
      end

c============================================================================

      subroutine galactic_equatorial (Dir,Equinox,l,b, al,de)
c  If Dir: galactic -> equatorial coordinates transformation, if
c   .not.Dir: equatorial -> galactic transformation.
      implicit NONE
      real*8 pi, e,fa,fl, e2,fa2,fl2, e1,fa1,fl1, Equinox,l,b, al,de, w
      parameter ( pi = 3.141592653589793115997963468544185161591D+00 )
c  Parameters for Equinox=B1950.0:
      parameter ( e1 = 62.6d0 /180.d0*pi )
      parameter ( fa1 = 282.25d0 /180.d0*pi )
      parameter ( fl1 = 33.0d0 /180.d0*pi )
c  Parameters for Equinox=J2000.0:
      parameter ( e2 = 62.8717d0 /180.d0*pi )
      parameter ( fa2 = 282.8596d0 /180.d0*pi )
      parameter ( fl2 = 32.93192d0 /180.d0*pi )
      logical Dir
      
      if (Equinox .eq. 1950.0d0)  then
         e = e1
         fa = fa1
         fl = fl1
        elseif (Equinox .eq. 2000.0d0)  then
         e = e2
         fa = fa2
         fl = fl2
        else
         write (*,*) ' In galactic_equatorial only Equinox=1950.0'//
     #               ' or 2000.0 are supported!'
         stop         
        end if
         

      if (Dir)  then
         w = datan2(dsin(l-pi/2.d0-fl),(dcos(l-pi/2.d0-fl)*cos(e)-
     #        dtan(b)*sin(e)))
         al = (w+fa-1.5d0*pi)
         de = dasin(sin(b)*cos(e)+cos(b)*sin(e)*cos(l-pi/2.d0-fl))
c  al - within interval [0...2*pi]
         al = al  - 2*pi*int(al/(2*pi))
         if (al .lt. 0.d0)  al = al + 2*pi
        else
         l = datan2(dcos(de)*dsin(al-fa)*dcos(e)+dsin(de)*dsin(e),
     #              dcos(de)*dcos(al-fa)) + fl
c  l - within interval [0...2*pi]
         l = l  - 2*pi*int(l/(2*pi))
         if (l .lt. 0.d0)  l = l + 2*pi
         b = dasin(dsin(de)*dcos(e)-dcos(de)*dsin(al-fa)*dsin(e))         
        end if

      return
      end
c============================================================================

      subroutine Extract_numbers (String, nn,Numbers)
c  Subroutine  to extract 1, 2 or 3 numbers
c  from the String, and to put them into array Numbers(1..nn).
      implicit NONE
      integer nn,i,i1,i2
      real*8 Numbers(3)
      character*(*) String, s*1
      logical st

      st = .false.
      nn = 0

      do i=1,len(string)
         s = string(i:i)
         if(s.eq.'0'.or.s.eq.'1'.or.s.eq.'2'.or.s.eq.'3'.or.s.eq.'4'.or.
     #      s.eq.'5'.or.s.eq.'6'.or.s.eq.'7'.or.s.eq.'8'.or.s.eq.'9'.or.
     #      s.eq.'+'.or.s.eq.'-'.or.s.eq.'.'.or.s.eq.'e'.or.s.eq.'E'.or.
     #      s.eq.'d'.or.s.eq.'D')  then
            if (.not.st)  i1 = i
            st = .true.
            else
            if (st) then
                  i2 = i-1
                  if (nn .lt. 3)  then
                     nn = nn + 1
                     read (string(i1:i2),*) Numbers(nn)
                     end if
                  st = .false.
               end if
            end if
         end do

      if (nn .eq. 0)  then
         write (*,*) 'There are no numbers in the input string!'
         end if

      end
c============================================================================

      function Alpha (Numbers,nn)
c  The function transforms three different formats of the equatorial coordinate
c  alpha into degrees:
c   nn=1: dgr. -> dgr. (no change); {dgr.=Numbers(1)}
c   nn=2: hh mm.m -> dgr.; {hh=Numbers(1), mm.m=Numbers(2)}
c   nn=3: hh mm ss.s -> dgr. {hh=Numbers(1), mm=Numbers(2), ss.s=Numbers(3)}
c  The result is returned in Alpha.
      implicit NONE
      integer nn
      real*8 Numbers(nn), Alpha

      if (nn.gt.3 .or. nn.lt.1)  then
         write (*,*) ' Wrong nn in Alpha!'
         stop
         end if

      if (nn .eq. 1)  then
         Alpha = Numbers(1)
        else if (nn .eq. 2)  then
         Alpha = 15.d0 * (Numbers(1) + Numbers(2)/60.d0)
        else
         Alpha = 15.d0 * (Numbers(1) + Numbers(2)/60.d0 + 
     #                    Numbers(3)/3600.d0)
        end if

      return
      end
c============================================================================

      function Delta (Numbers,nn)
c  The function transforms three different formats of the equatorial coordinate
c  delta into degrees:
c   nn=1: dgr. -> dgr. (no change); {dgr.=Numbers(1)}
c   nn=2: dd mm.m -> dgr.; {dd=Numbers(1), mm.m=Numbers(2)}
c   nn=3: dd mm ss.s -> dgr. {dd=Numbers(1), mm=Numbers(2), ss.s=Numbers(3)}
c  The result is returned in Delta.
      implicit NONE
      integer nn
      real*8 Numbers(nn), Delta

      if (nn.gt.3 .or. nn.lt.1)  then
         write (*,*) ' Wrong nn in Delta!'
         stop
         end if

      if (nn .eq. 1)  then
         Delta = Numbers(1)
        else if (nn .eq. 2)  then
         Delta = dsign(1.d0,Numbers(1)) * (dabs(Numbers(1)) + 
     #           Numbers(2)/60.d0)
        else
         Delta = dsign(1.d0,Numbers(1)) * (dabs(Numbers(1)) + 
     #                 Numbers(2)/60.d0 + Numbers(3)/3600.d0)
        end if

      return
      end
c============================================================================


      INTEGER FUNCTION strlen_trail (STRING)
C
C Return the length of a string less any trailing blanks
C
C
C   STRING      CH*(*)  input   String to be measured
      implicit NONE
      CHARACTER*(*)     STRING
      INTEGER           I, N

      N = LEN (STRING)
      DO 1 I = N,1,-1
         IF(STRING(I:I).NE.' ') GOTO 2
    1 CONTINUE
    2 STRLEN_trail = I
      END
c--------------------------------------------------------------------------


      INTEGER FUNCTION strlen_lead (STRING)
C
C Return the length of a string less any leading blanks
C
C
C   STRING      CH*(*)  input   String to be measured

      implicit NONE
      CHARACTER*(*)     STRING
      INTEGER           I, N

      N = LEN (STRING)
      DO 1 I = 1,N
         IF(STRING(I:I).NE.' ') GOTO 2
    1 CONTINUE
    2 STRLEN_lead = N-I+1
      return
      END
c-------------------------------------------------------------------------
